{
Copyright (c) 2011, Carl-Philip Hänsch

Permission to use, copy, modify, and/or distribute this software for any purpose with or
without fee is hereby granted, provided that the above copyright notice and this permission
notice appear in all copies and for each binary release, a linux version and the source code
is developed and released.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY
DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
OR PERFORMANCE OF THIS SOFTWARE.
}

unit ooenet;
{$mode objfpc}{$H+}
interface
uses generalhelpers, enet, sysutils, classes;

 type

  TEnetPacketFlags = set of (pfReliable, pfUnsequenced, pfNoAllocate);

  { TEnetHost }

  TEnetHostPeer = class;

  {
   Der TEnetHost ist der Server und öffnet an einem bestimmten Port.
   Er kann den Clienten Nachrichten senden und diese müssen dann geflusht werden
  }
  TEnetHost = class (TTreeComponent)
  private
    fhost: PENetHost;
    fclients: TFPList;
    function getclient(index: integer): TEnetHostPeer;
    procedure setclient(index: integer; const AValue: TEnetHostPeer);
    procedure removePeer(peer: PEnetPeer);
  public
    //
    constructor create(aowner: TTreeComponent; port: integer;
                connections: integer = 512; downstream: integer = 0;
                upstream: integer = 0); reintroduce; // Erstellt den Host am Port
    destructor destroy; override;
    procedure flush;  // sendet alle zu sendenden Nachrichten
    procedure handleReceives(timeout: integer); // empfängt alle Nachrichten
              //und bearbeitet sie

    function count: integer; // Wieviele Clients hat er
    property client[index: integer]: TEnetHostPeer read getclient
             write setclient; default; // Clients adressieren
    property host:PENetHost read fhost;
    
    // zu überladende Funktionen
    function allocateClient: TEnetHostPeer; virtual; abstract; // Soll einen neuen Clienten zurückgeben
    function allownewClient: boolean; virtual; abstract; // sagt, ob neue Verbindungen erlaubt sind
  end;

  { TEnetPeer }

  {
   Der TEnetPeer ist die Gemeinsamkeit zwischen einem Clienten auf ClientSide
   und einem Client-Handle auf der Server-Side. Von dieser Klasse nicht erben!
  }
  TEnetPeer = class (TTreeComponent)
  private
    fpeer: PENetPeer;
  public
    destructor destroy; override;

    procedure send(s: string; flags: TEnetPacketFlags); // sendet an den Peer

    // zu überladen
    procedure receive(s: string); virtual; abstract; // wenn empfangen wird
  end;

  { TEnetHostPeer }

  {
   Der TEnetHostPeer ist ein Client auf der Serverside, der sich angemeldet hat
  }
  TEnetHostPeer = class (TEnetPeer)
  private
    fhost: TEnetHost;
  public
    ip: string;

    function getName: string;
    procedure kick; // Schließt die Verbindung zum Client
    
    // zu überladen
    procedure init; virtual; abstract; // Callback wenn der Client verbunden ist
    procedure deinit; virtual; abstract; // Callback wenn der Client geht
  end;

  { TEnetClient }

  {
   Der TEnetClient ist der Client auf ClientSide und verbindet sich mit einem
   Server. Er kann sich immer wieder verbinden.
  }
  TEnetClient = class (TEnetPeer)
  private
    fhost: PENetHost;
    fconnected: boolean;
  public
    function connect(address: string; port: integer; timeout: integer): boolean;
    procedure disconnect; // Meldet den Client ab
    procedure flush; // Sendet alle gesendeten Nachrichten ab
    procedure handleReceives(timeout: integer); // fängt an, Nachrichten abzuarbeiten
    function isConnected: boolean; // ob er verbunden ist

    // zu überladen
    procedure connectionLoss; virtual; abstract; // Callback bei Verbindungsabbruch
  end;

implementation

var locker: TRTLCriticalSection;
procedure enet_lock;
begin
 EnterCriticalsection(locker);
end;
procedure enet_unlock;
begin
 LeaveCriticalsection(locker);
end;

function convertFlags(flags: TEnetPacketFlags): cardinal;
begin
  result:=0;
  if pfReliable in flags then result:=result or ENET_PACKET_FLAG_RELIABLE;
  if pfUnsequenced in flags then result:=result or ENET_PACKET_FLAG_UNSEQUENCED;
  if pfNoAllocate in flags then result:=result or ENET_PACKET_FLAG_NO_ALLOCATE;
end;

{ TEnetHost }

function TEnetHost.getclient(index: integer): TEnetHostPeer;
begin
  result:=TEnetHostPeer(fclients[index]);
end;

procedure TEnetHost.setclient(index: integer; const AValue: TEnetHostPeer);
begin
  fclients[index]:=AValue;
end;

constructor TEnetHost.create(aowner: TTreeComponent; port: integer;
  connections: integer; downstream: integer; upstream: integer);
var adr: ENetAddress;
begin
  fclients:=TFPList.Create;
  adr.host:=ENET_HOST_ANY;
  adr.port:=port;
  enet_lock;
  fhost:=enet_host_create(@adr, connections, {$IFNDEF ENET12}1,{$ENDIF} downstream, upstream);
  enet_unlock;
  inherited create(aowner);
end;

procedure TEnetHost.removePeer(peer: PEnetPeer);
var i: integer;
begin
 i:=fclients.count-1;
 while i>=0 do begin
   if TEnetHostPeer(fclients[i]).fpeer=peer then begin
    TEnetHostPeer(fclients[i]).deinit;
    TEnetHostPeer(fclients[i]).fpeer:=nil;
    TEnetHostPeer(fclients[i]).Free;
    fclients.Delete(i);
    i:=-1;
   end else dec(i);
 end;
end;

destructor TEnetHost.destroy;
var i: integer;
begin
  for i:=count-1 downto 0 do client[i].kick;
  flush; // erst die disconnect-Nachrichten senden
  
  inherited destroy;
  enet_lock;
  if fhost<>nil then enet_host_destroy(fhost);
  enet_unlock;
  fclients.free;
end;

procedure TEnetHost.flush;
begin
  enet_lock;
  enet_host_flush(fhost);
  enet_unlock;
end;

function iptostring(nr: cardinal): string;
begin
 result:=inttostr((nr shr 00) and $FF)+'.'+
         inttostr((nr shr 08) and $FF)+'.'+
         inttostr((nr shr 16) and $FF)+'.'+
         inttostr((nr shr 24) and $FF);
end;

procedure TEnetHost.handleReceives(timeout: integer);
var event: ENetEvent;
    neu: TEnetHostPeer;
    i: integer;
    s: string;
begin
  enet_lock;
  while enet_host_service(fhost, @event, timeout)>0 do begin
   enet_unlock;
   case event.typ of
    ENET_EVENT_TYPE_CONNECT:begin
       if allownewClient then begin
         neu:=allocateClient; // user-Call: nur erstellen
         neu.fpeer:=event.peer;
         neu.fhost:=self;
         neu.ip:=iptostring(neu.fpeer^.address.host);
         neu.init; // hier kann er senden und so
         fclients.add(neu);
       end else begin
         enet_lock;
         enet_peer_disconnect(event.peer, 0);
         enet_unlock;
       end;
    end;
    ENET_EVENT_TYPE_DISCONNECT:begin
       removePeer(event.peer);

    end;
    ENET_EVENT_TYPE_RECEIVE:begin
      if event.packet^.dataLength>0 then begin
       setlength(s, event.packet^.dataLength);
       move((event.packet^.data)^, s[1], event.packet^.dataLength);
       for i:=fclients.Count-1 downto 0 do if TEnetHostPeer(fclients[i]).fpeer=event.peer then
          TEnetHostPeer(fclients[i]).receive(s);
       enet_lock;
       enet_packet_destroy(event.packet);
       enet_unlock;
      end;
    end;
   end;
   enet_lock;
  end;
  enet_unlock;
end;

function TEnetHost.count: integer;
begin
 result:=fclients.Count;
end;

{ TEnetHostPeer }

destructor TEnetPeer.destroy;
begin
  inherited destroy;
  enet_lock;
  if fpeer<>nil then enet_peer_reset(fpeer);
  fpeer:=nil;
  enet_unlock;
end;

procedure TEnetPeer.send(s: string; flags: TEnetPacketFlags);
begin
  if fpeer=nil then exit;
  enet_lock;
  enet_peer_send_string(fpeer, s, convertFlags(flags));
  enet_unlock;
end;

function TEnetHostPeer.getName: string;
begin
  if fpeer=nil then exit('disconnected');
  enet_lock;
  result:=enet_address_string(fpeer^.address);
  enet_unlock;
end;

procedure TEnetHostPeer.kick;
begin
  enet_lock;
  enet_peer_disconnect(fpeer, 0);
  enet_unlock;
  fhost.removePeer(fpeer);
end;

{ TEnetClient }

function TEnetClient.connect(address: string; port: integer; timeout: integer): boolean;
var ip: ENetAddress;
    event: ENetEvent;
    i: integer;
begin
  if fhost<>nil then disconnect;
  enet_lock;
  enet_address_set_host(@ip, pchar(address));
  ip.port:=port;
  fhost:=enet_host_create(nil, 1, {$IFNDEF ENET12}1,{$ENDIF} 0, 0); // Client-Socket erstellen
  fpeer:=enet_host_connect(fhost, @ip, {$IFNDEF ENET12}1,{$ENDIF} 0); // mit dem Server verbinden

  for i:=0 to timeout div 10 do begin
   if enet_host_service(fhost, @event, 5) > 0 then begin // timeout Millisekunden lang verbinden
    enet_unlock;
    if event.typ=ENET_EVENT_TYPE_CONNECT then begin // ja er ist drin
     result:=true;
     fpeer:=event.peer;
     fconnected:=true;
     exit;
    end;
   end else begin
    enet_unlock;
    sleep(5);
   end;
   enet_lock;
  end;
  // sonst: Alles wieder freigeben
  result:=false;
  fconnected:=false;
  enet_peer_reset(fpeer);
  fpeer:=nil;
  enet_host_destroy(fhost);
  fhost:=nil;
  enet_unlock;
end;

procedure TEnetClient.disconnect;
begin
  fconnected:=false;
  if fpeer<>nil then begin
   enet_peer_disconnect(fpeer, 0); // ordentlich abmelden
   flush;
   enet_peer_reset(fpeer);
   fpeer:=nil;
  end;
  if fhost<>nil then begin
    enet_lock;
    enet_host_destroy(fhost);
    fhost:=nil;
    enet_unlock;
  end;
end;

procedure TEnetClient.flush;
begin
  if not isConnected then exit;
  enet_lock;
  enet_host_flush(fhost);
  enet_unlock;
end;

procedure TEnetClient.handleReceives(timeout: integer);
var s: string;
    event: ENetEvent;
begin
 if not isConnected then exit;
 enet_lock;
 while enet_host_service (fhost, @event, timeout) > 0 do begin
  enet_unlock;
  case event.typ of
   ENET_EVENT_TYPE_CONNECT: begin
    raise Exception.create('This should not happen: connect after connecting');
   end;
   ENET_EVENT_TYPE_DISCONNECT: begin
    disconnect;
    connectionLoss; // User-Event
   end;
   ENET_EVENT_TYPE_RECEIVE: begin
     if event.packet^.dataLength>0 then begin
      setlength(s, event.packet^.dataLength);
      move((event.packet^.data)^, s[1], event.packet^.dataLength);
      receive(s); // Nachricht kommt an
      enet_lock;
      enet_packet_destroy(event.packet);
      enet_unlock;
     end;
   end;
  end;
  if not isConnected then exit;
  enet_lock;
 end;
 enet_unlock;

end;

function TEnetClient.isConnected: boolean;
begin
  result:=fconnected;
end;

initialization
 InitCriticalSection(locker);
finalization
 DoneCriticalsection(locker);

end.
